#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include "path.h"
#include "option.h"
#include "mesh.h"
#include "blackScholes.h"
#include "control3.h"
#include "util.h"


double weight(meshData *md, optionData *od, pathData *pd, int time,
              double *prices, int toPath){

  return( pathTransProb(pd,prices,pd->price[time+1][toPath]) /
          md->distribution[time+1][toPath]);
}


double controlledMeshEstimate0(meshData *md, optionData *od, pathData *pd,
                               int time, double *prices, double *variance){
  /* Mesh estimate without inner controls.
   */

  int toPath;
  double f,w,S1,S2,S3,S4,S5,S6;

  S1 = 0;
  S2 = 0;
  S3 = 0;
  S4 = 0;
  S5 = 0;
  S6 = 0;
  for(toPath=0; toPath<md->meshPaths; toPath++){
    f = md->value[time+1][toPath];
    w = weight(md,od,pd,time,prices,toPath);
    S1 += w;
    S2 += w*w;
    S3 += w*f;
    S4 += w*w*f;
    S6 += w*w*f*f;
  }
  *variance = (1/(S1*S1))*(S6 - 2*S3/S1*S4 + (S3/S1)*(S3/S1)*S2);

  return(S3/S1);
}


double controlledMeshEstimate3(meshData *md, optionData *od, pathData *pd,
                               int time, double *prices, double *variance){
  /* Returns a mesh estimate using inner control 3.
   * Sets variance to the estimated variance.
   */

  int toPath, max1, max2, i;
  double S1, S2, S7, S8, S9 ,S10, S11, S12, S13, w, f, g, G, maxval1, maxval2,
    maxPrice, beta, S14, S15, S16;

  S1 = 0;
  S2 = 0;
  S7 = 0;
  S8 = 0;
  S9 = 0;
  S10 = 0;
  S11 = 0;
  S12 = 0;
  S13 = 0;
  S14 = 0;
  S15 = 0;
  S16 = 0;

  /* At the given time find the max prices */

  maxval1 = 0.0;
  maxval2 = 0.0;
  max1 = dWhichMax(pd->underliers, prices);
  maxval1 = prices[max1];

  for(i=0; i<pd->underliers;i++){
    if(i != max1 && prices[i] > maxval2){
      max2 = i;
      maxval2 = prices[i];
    }
  }

  G = exp(-time*od->riskFreeRate)* 
    maxOf2Risk(maxval1, maxval2, od->strikePrice, 1.0,
               sqrt(pd->covariance[max1][max1]), 
               sqrt(pd->covariance[max2][max2]), 
               od->riskFreeRate, 
               pd->dividendRate[max1], 
               pd->dividendRate[max2],
               pd->covariance[max1][max2]);

  for(toPath = 0; toPath < md->meshPaths; toPath++){
    w = weight(md,od,pd,time,prices,toPath);

    f = md->value[time+1][toPath];

    if( pd->price[time+1][toPath][max1] >= pd->price[time+1][toPath][max2]){
      maxPrice = pd->price[time+1][toPath][max1];
    }else{
      maxPrice = pd->price[time+1][toPath][max2];
    }
    if(maxPrice > od->strikePrice){
      g = exp(-(time+1)*od->riskFreeRate)*(maxPrice - od->strikePrice);
    }else{
      g= 0;
    }

    S1 += w;
    S2 += w*w;
    S7 += w*f;
    S8 += w*(g-G);
    S9 += w*w*f;
    S10 += w*w*(g-G);
    S11 += w*w*f*f;
    S12 += w*w*f*(g-G);
    S13 += w*w*(g-G)*(g-G);
    S14 += w*f*(g-G);
    S15 += w*(g-G)*(g-G);
    S16 += w*f*f;
  }

  /* New derivation */
  beta = (S12*S1*S1 - (S8*S9 + S7*S10)*S1 + S2*S7*S8)/   
    (S13*S1*S1 - 2*S8*S10*S1 + S2*S8*S8);   

  *variance = 1/(S1*S1)*((S11 -2*beta*S12 + beta*beta*S13)  
                         -(2/S1)*(S7-beta*S8)*(S9-beta*S10)  
                         +(S2/(S1*S1))*(S7*S7 -2*beta*S7*S8 + beta*beta*S8*S8)  
                         );  


  /* Old derivation */
  /*      beta = ((S14/S1) - (S7/S1)*(S8/S1))/   */
  /*             ((S15/S1) - (S8/S1)*(S8/S1));   */


/*      *variance = S16/S1 - (S7/S1)*(S7/S1)  */
/*                 -2*beta*(S14/S1 - (S7/S1)*(S8/S1))  */
/*                 +beta*beta*((S15/S1) - (S8/S1)*(S8/S1));  */

  return( (S7-beta*S8)/S1);

}





void meshDataFill(meshData *md, optionData *od, pathData *pd,
                  int whichControl){

  int time, path, toPath, fromPath;
  double var;

  for(time=pd->times-2; time>=0; time--){
    for(toPath=0; toPath<md->meshPaths; toPath++){
      md->distribution[time+1][toPath] = 0;
      for(fromPath=0; fromPath<md->meshPaths;fromPath+=pd->clusterSize){
        md->distribution[time+1][toPath] += pd->clusterSize*
          pathTransProb(pd, pd->price[time][fromPath],
                        pd->price[time+1][toPath]);
      }
    }
  }

  for(time=pd->times-2; time>0; time--){
    for(path=md->meshPaths-1; path>=0; path--){
      switch(whichControl){
      case(3):
        md->valueIfHeld[time][path] =
          controlledMeshEstimate3(md,od,pd,time,pd->price[time][path],&var);
        md->variance[time][path] = var;
        break;
      default:
        md->valueIfHeld[time][path] = 
          controlledMeshEstimate0(md,od,pd,time,pd->price[time][path],&var);
        md->variance[time][path] = var;
        break;
      }

      if(md->valueIfHeld[time][path] > od->payoff[time][path]){
        md->value[time][path] = md->valueIfHeld[time][path];
      }else{
        md->value[time][path] = od->payoff[time][path];
      }
    }
  }

  switch(whichControl){
  case(3):
    md->valueIfHeld[0][0] =
      controlledMeshEstimate3(md,od,pd,0,pd->price[time][0],&var);
    md->variance[0][0] = var;
    break;
  default:
    md->valueIfHeld[0][0] = 
      controlledMeshEstimate0(md,od,pd,0,pd->price[time][0],&var);
    md->variance[0][0] = var;
    break;
  }

  if(md->valueIfHeld[0][0] > od->payoff[0][0]){
    md->value[0][0] = md->valueIfHeld[0][0];
  }else{
    md->value[0][0] = od->payoff[0][0];
  }
}




double innerControlledPathEstimate(meshData *md, optionData *od, pathData *pd,
                                   int whichControl){
  int time, underlier, toPath, count,max1,max2,i;
  double **path, **antiPath, antiGrowth, *payoff, *antiPayoff, *valueIfHeld,
    *antiValueIfHeld, value, antiValue,  delta, euroVal,
    maxval1, maxval2, var;
  
  allocate2dDouble(&path, pd->times, pd->underliers);
  allocate2dDouble(&antiPath,  pd->times, pd->underliers);
  allocate1dDouble(&payoff, pd->times);
  allocate1dDouble(&antiPayoff, pd->times);
  allocate1dDouble(&valueIfHeld, pd->times);
  allocate1dDouble(&antiValueIfHeld, pd->times);
  
  
  value = 0;
  antiValue = 0;
  
  for(count=1; count<=pd->paths; count++){
    pathPair(pd,path,antiPath);

    for(time=0; time < pd->times; time++){
      payoff[time] = exp(-time*od->riskFreeRate)*
        optionPayoff(od, pd, path[time]);
      antiPayoff[time] = exp(-time*od->riskFreeRate)*
        optionPayoff(od, pd, antiPath[time]);
    }

    valueIfHeld[0] = dMax(0,payoff[0]);
    antiValueIfHeld[0] = dMax(0,antiPayoff[0]);
    
    for(time=0; time < pd->times; time++){
      /* We don't need to compute value if held on the last timeslice */
      if(time == pd->times-1){
        if(payoff[time] > valueIfHeld[time]){
          value += payoff[time];
          break;
        }
      }else{
        /* Skip point if payoff not > 0 */
        if(payoff[time] > 0 ){
          max1 = dWhichMax(pd->underliers, path[time]); 
          euroVal = exp(-time*od->riskFreeRate)*
            euroCall(path[time][max1],
                     od->strikePrice,
                     od->riskFreeRate,
                     sqrt(pd->covariance[max1][max1]),
                     pd->times-1 - time,
                     pd->dividendRate[max1]);
          /* Skip point if payoff not > 1d european */
          if(payoff[time] > euroVal){
            maxval1 = path[time][max1];
            maxval2 = 0;
            for(i = 0; i<pd->underliers; i++){
              if(i != max1 && path[time][i] > maxval2){
                max2 = i;
                maxval2 = path[time][i];
              }
            } 
            euroVal = exp(-time*od->riskFreeRate)* 
              maxOf2Risk(maxval1, maxval2, od->strikePrice,
                         pd->times-1 - time,
                         sqrt(pd->covariance[max1][max1]), 
                         sqrt(pd->covariance[max2][max2]), 
                         od->riskFreeRate, 
                         pd->dividendRate[max1], 
                         pd->dividendRate[max2],
                         pd->covariance[max1][max2]);
            /* Skip point if payoff not > 2d european */
            if(payoff[time] > euroVal){
              switch(whichControl){
              case(3):
                valueIfHeld[time] =
                  controlledMeshEstimate3(md,od,pd,time,path[time],&var);
                break;
              default:
                valueIfHeld[time] = 
                  controlledMeshEstimate0(md,od,pd,time,path[time],&var);
                break;
              }

              if(payoff[time] > valueIfHeld[time]){
                value += payoff[time];
                break;
              }
            }
          }
        }
      }
    }

    for(time=0; time < pd->times; time++){
      /* We don't need to compute value if held on the last timeslice */
      if(time == pd->times-1){
        if(antiPayoff[time] > antiValueIfHeld[time]){
          antiValue += antiPayoff[time];
          break;
        }
      }else{
        /* Skip point if antiPayoff not > 0 */
        if(antiPayoff[time] > 0 ){
          max1 = dWhichMax(pd->underliers, antiPath[time]);
          euroVal = exp(-time*od->riskFreeRate)*
            euroCall(antiPath[time][max1],
                     od->strikePrice,
                     od->riskFreeRate,
                     sqrt(pd->covariance[max1][max1]),
                     pd->times-1 - time,
                     pd->dividendRate[max1]);
          /* Skip point if antiPayoff not > 1d european */
          if(antiPayoff[time] > euroVal){
            maxval1 = antiPath[time][max1];
            maxval2 = 0;
            for(i = 0; i<pd->underliers; i++){
              if(i != max1 && antiPath[time][i] > maxval2){
                max2 = i;
                maxval2 = antiPath[time][i];
              }
            } 
            euroVal = exp(-time*od->riskFreeRate)* 
              maxOf2Risk(maxval1, maxval2, od->strikePrice,
                         pd->times-1 - time,
                         sqrt(pd->covariance[max1][max1]), 
                         sqrt(pd->covariance[max2][max2]), 
                         od->riskFreeRate, 
                         pd->dividendRate[max1], 
                         pd->dividendRate[max2],
                         pd->covariance[max1][max2]);
            /* Skip point if antiPayoff not > 2d european */
            if(antiPayoff[time] > euroVal){
              switch(whichControl){
              case(3):
                antiValueIfHeld[time] =
                  controlledMeshEstimate3(md,od,pd,time,antiPath[time],&var);
                break;
              default:
                valueIfHeld[time] = 
                  controlledMeshEstimate0(md,od,pd,time,antiPath[time],&var);
                break;
              }
              
              if(antiPayoff[time] > antiValueIfHeld[time]){
                antiValue += antiPayoff[time];
                break;
              }
            }
          }
        }
      }
    }

  }

  value /= pd->paths;
  antiValue /= pd->paths;
  
  deallocate2dDouble(&path,  pd->times,
                     pd->underliers); 
  deallocate2dDouble(&antiPath,  pd->times,
                     pd->underliers); 
  deallocate1dDouble(&payoff, pd->times); 
  deallocate1dDouble(&antiPayoff, pd->times); 
  deallocate1dDouble(&valueIfHeld, pd->times); 
  deallocate1dDouble(&antiValueIfHeld, pd->times); 

  return(0.5*(value+antiValue));  
}
